unit ListO;

  (*--------------------------------------------------------------------------*)
  (*                                                                          *)
  (* TListO:     Object-based (as opposed to pointers) version of TList.      *)
  (* Version:    990904.1623                                                  *)
  (*                                                                          *)
  (* Copyright:  Jim O'Flaherty, Jr. - 1999                                   *)
  (* Created On: 1999 Sep 02                                                  *)
  (*                                                                          *)
  (*--------------------------------------------------------------------------*)

  interface

    uses
      Classes;

    type
      TListOSortCompare = function (oWhat1, oWhat2: TObject): Integer of object;
          //Funtion provided should return:
          //   < 0 (-1) when oWhat1 < oWhat2
          //   = 0      when oWhat1 = oWhat2
          //   > 0 (1)  when oWhat1 > oWhat2
      TListO = class
        protected
          lsContent: TList;
          function GetCapacity: Integer; virtual;
          procedure SetCapacity(iWhat: Integer); virtual;
          function GetCount: Integer; virtual;
          procedure SetCount(iWhat: Integer); virtual;
          function GetItems(iWhat: Integer): TObject; virtual;
          procedure SetItems(iWhat: Integer; oWhat: TObject); virtual;
          function GetList: PPointerList; virtual;
        public
          constructor Create; virtual;
          destructor Destroy; override;

          function Add(oWhat: TObject): Integer; virtual;
              //Append the entry.
          procedure Clear; virtual;
          procedure Delete(iWhat: Integer); virtual;
          procedure Error(const Msg: string; Data: Integer); virtual;
          procedure Exchange(iWhat1, iWhat2: Integer); virtual;
          function Expand: TListO; virtual;
          function First: TObject; virtual;
          function IndexOf(oWhat: TObject): Integer; virtual;
          function IndexOfInRange(oWhat: TObject; iStartingAt, iEndingAt: Integer): Integer; virtual;
              //New
          procedure Insert(iWhat: Integer; oWhat: TObject); virtual;
          function Last: TObject; virtual;
          function LastIndex: Integer; virtual;
              //New
          procedure Move(iWhat1, iWhat2: Integer); virtual;
          procedure Pack; virtual;
          function Remove(oWhat: TObject): Integer; virtual;
          procedure Sort(lsoscWhat: TListOSortCompare); virtual;
          class procedure QuickSort(SortList: PPointerList; L, R: Integer;
              SCompare: TListOSortCompare);

          property Capacity: Integer read GetCapacity write SetCapacity;
          property Count: Integer read GetCount write SetCount;
          property Items[iWhat: Integer]: TObject read GetItems write SetItems;
          property List: PPointerList read GetList;
        end;

  implementation

//protected
    function TListO.GetCapacity: Integer;
      begin
      Result := lsContent.Capacity;
      end;

    procedure TListO.SetCapacity(iWhat: Integer);
      begin
      lsContent.Capacity := iWhat;
      end;

    function TListO.GetCount: Integer;
      begin
      Result := lsContent.Count;
      end;

    procedure TListO.SetCount(iWhat: Integer);
      begin
      lsContent.Count := iWhat;
      end;

    function TListO.GetItems(iWhat: Integer): TObject;
      begin
      Result := TObject(lsContent.Items[iWhat]);
      end;

    procedure TListO.SetItems(iWhat: Integer; oWhat: TObject);
      begin
      lsContent.Items[iWhat] := Pointer(oWhat);
      end;

    function TListO.GetList: PPointerList;
      begin
      Result := lsContent.List;
      end;
//public
    constructor TListO.Create;
      begin
      lsContent := TList.Create;
      end;

    destructor TListO.Destroy;
      begin
      lsContent.Free;
      inherited Destroy;
      end;

    function TListO.Add(oWhat: TObject): Integer;
      begin
      Result := lsContent.Add(Pointer(oWhat));
      end;

    procedure TListO.Clear;
      begin
      lsContent.Clear;
      end;

    procedure TListO.Delete(iWhat: Integer);
      begin
      lsContent.Delete(iWhat);
      end;

    procedure TListO.Error(const Msg: string; Data: Integer);
      begin
      lsContent.Error(Msg, Data);
      end;

    procedure TListO.Exchange(iWhat1, iWhat2: Integer);
      begin
      lsContent.Exchange(iWhat1, iWhat2);
      end;

    function TListO.Expand: TListO;
      begin
      lsContent.Expand;
      Result := Self;
      end;

    function TListO.First: TObject;
      begin
      Result := TObject(lsContent.First);
      end;

    function TListO.IndexOf(oWhat: TObject): Integer;
      begin
      Result := lsContent.IndexOf(Pointer(oWhat));
      end;

    function TListO.IndexOfInRange(oWhat: TObject; iStartingAt, iEndingAt: Integer): Integer;
      var
        i: Integer;
      begin
      if (not ((iStartingAt >= 0) and (iStartingAt <= iEndingAt))) then
        Error('TListO.IndexOfInRange: iStartingAt not in valid range of 0..iEndingAt', iStartingAt);
      if (not ((iEndingAt >= iStartingAt) and (iEndingAt <= LastIndex))) then
        Error('TListO.IndexOfInRange: iEndingAt not in valid range of iStartingAt..LastIndex', iEndingAt);
      i := iStartingAt;
      while (i <= iEndingAt) and (Items[i] <> oWhat) do
        Inc(i);
      Result := lsContent.IndexOf(Pointer(oWhat));
      end;

    procedure TListO.Insert(iWhat: Integer; oWhat: TObject);
      begin
      lsContent.Insert(iWhat, Pointer(oWhat));
      end;

    function TListO.Last: TObject;
      begin
      Result := TObject(lsContent.Last);
      end;

    function TListO.LastIndex: Integer;
      begin
      Result := lsContent.Count - 1;
      end;

    procedure TListO.Move(iWhat1, iWhat2: Integer);
      begin
      lsContent.Move(iWhat1, iWhat2);
      end;

    procedure TListO.Pack;
      begin
      lsContent.Pack;
      end;

    function TListO.Remove(oWhat: TObject): Integer;
      begin
      Result := lsContent.Remove(Pointer(oWhat));
      end;

    procedure TListO.Sort(lsoscWhat: TListOSortCompare);
      begin
      //Note: Unable to use lsContent.Sort as the signature *requires* a unit
      //      procedure and cannot use a method pointer!
      if (List <> nil) and (Count > 0) then
        QuickSort(List, 0, LastIndex, lsoscWhat);
      end;

//Note: Copied this from unit Classes and then modified it to work with an
//      object method pointer.
    class procedure TListO.QuickSort(SortList: PPointerList; L, R: Integer;
        SCompare: TListOSortCompare);
      var
        I, J: Integer;
        P, T: Pointer;
      begin
      repeat
        I := L;
        J := R;
        P := SortList^[(L + R) shr 1];
        repeat
          while SCompare(TObject(SortList^[I]), TObject(P)) < 0 do
            Inc(I);
          while SCompare(TObject(SortList^[J]), TObject(P)) > 0 do
            Dec(J);
          if I <= J then
            begin
            T := SortList^[I];
            SortList^[I] := SortList^[J];
            SortList^[J] := T;
            Inc(I);
            Dec(J);
            end;
        until I > J;
        if L < J then
          QuickSort(SortList, L, J, SCompare);
        L := I;
      until I >= R;
      end;

  end.
